-- 
-- Data Model for storing metadata about acs_objects for use in a
-- web interface to building new data models
-- 
-- @author Kevin Scaldeferri
-- @creation-date 2000-12-21
-- @cvs-id $Id: rapid-development-create.sql,v 1.1 2001/06/03 00:54:27 dm Exp $
-- 

--
-- I don't see any real reason to use acs_objects here.
-- This way, this tool can operate independently of the 
-- rest of the ACS, which might be handy.
--

create sequence rd_table_id_seq;

create table rd_tables (
    table_id           integer 
                       constraint rd_tables_pk
                       primary key,
    table_name         varchar2(30) 
                       constraint rd_tables_table_name_nn
                       not null 
                       constraint rd_tables_table_name_un
                       unique,
    -- columns for acs_object_types
    id_column          varchar2(30)
                       constraint rd_tables_id_column_nn
		       not null,
    object_type        varchar2(100) 
                       constraint rd_tables_object_type_nn
                       not null 
                       constraint rd_tables_object_type_un
                       unique,
    supertype          varchar2(100),
    package_name       varchar2(30) 
                       constraint rd_tables_package_name_nn
                       not null 
                       constraint rd_tables_package_name_un
                       unique,
    pretty_name        varchar2(100) 
                       constraint rd_tables_pretty_name_nn
                       not null 
                       constraint rd_tables_pretty_name_un
                       unique,
    pretty_plural      varchar2(100) 
                       constraint rd_tables_pretty_plural_nn
                       not null 
                       constraint rd_tables_pretty_plural_un
                       unique,
    -- to be used if we need to shorten a constraint name
    abbrev             varchar2(13),
    -- description of what this table is for
    description        varchar2(4000)
);

create sequence rd_column_id_seq;

create table rd_columns (
    column_id           integer 
                        constraint rd_columns_pk
                        primary key,
    table_id            integer 
                        constraint rd_columns_table_id_fk
                        references rd_tables 
                        on delete cascade
                        constraint rd_columns_table_id_nn
                        not null,
    column_name         varchar2(30) 
                        constraint rd_columns_column_name_nn
                        not null,
    -- can't put in a check constraint on datatype because
    -- varchars take a length
    datatype            varchar2(50) 
                        constraint rd_columns_datatype_nn
                        not null,
    -- acs_datatype?
    default_value       varchar2(200),
    attribute_p         char(1) default 't'
                        constraint rd_columns_attribute_p_ck
                        check (attribute_p in ('t','f')),
    -- to be used if we need to shorten a constraint name
    abbrev              varchar2(13),
    description         varchar2(4000),
    -- columns for attributes
    pretty_name         varchar2(100),
    pretty_plural       varchar2(100),
    constraint rd_cols_table_id_col_name_un 
    unique (table_id, column_name)
);


-- At the moment we are only going to support column constraints

create sequence rd_constraint_id_seq;

create table rd_constraints (
    constraint_id           integer
                            constraint rd_constraints_pk
                            primary key,
    column_id               integer
                            constraint rd_constraints_column_id_fk
                            references rd_columns 
                            on delete cascade
                            constraint rd_constraints_column_id_nn
                            not null,
    -- I'm not sure if we will actually support check constraints
    -- at the moment (other than booleans, which will probably
    -- be denoted in the datatype rather than here.)
    constraint_type         varchar2(20)
                            constraint rd_constraints_type_ck
                            check (constraint_type in (
                                'primary key',
                                'foreign key',
                                'unique',
                                'not null',
                                'check')),
    constraint_body         varchar2(500),
    constraint rd_constrnts_col_type_body_un
    unique (column_id, constraint_type, constraint_body)
);



create or replace package rapid_development
is

    function add_table (
        table_id       in rd_tables.table_id%TYPE default null,
        table_name     in rd_tables.table_name%TYPE default null,
        id_column      in rd_tables.id_column%TYPE default null,
        object_type    in rd_tables.object_type%TYPE,
        supertype      in rd_tables.supertype%TYPE default null,
        package_name   in rd_tables.package_name%TYPE default null,
        pretty_name    in rd_tables.pretty_name%TYPE,
        pretty_plural  in rd_tables.pretty_plural%TYPE default null,
        abbrev         in rd_tables.abbrev%TYPE default null,
        description    in rd_tables.description%TYPE default null
    ) return rd_tables.table_id%TYPE;

    procedure drop_table (
        table_id       in rd_tables.table_id%TYPE
    );

    procedure edit_table (
        table_id       in rd_tables.table_id%TYPE,
        table_name     in rd_tables.table_name%TYPE,
        id_column      in rd_tables.id_column%TYPE,
        object_type    in rd_tables.object_type%TYPE,
        supertype      in rd_tables.supertype%TYPE,
        package_name   in rd_tables.package_name%TYPE,
        pretty_name    in rd_tables.pretty_name%TYPE,
        pretty_plural  in rd_tables.pretty_plural%TYPE,
        abbrev         in rd_tables.abbrev%TYPE default null,
        description    in rd_tables.description%TYPE default null
    );

    function add_column (
        column_id       in rd_columns.column_id%TYPE default null,
        table_id        in rd_columns.table_id%TYPE,
        column_name     in rd_columns.column_name%TYPE,
        datatype        in rd_columns.datatype%TYPE,
        default_value   in rd_columns.default_value%TYPE default null,
        attribute_p     in rd_columns.attribute_p%TYPE default 't',
        pretty_name     in rd_columns.pretty_name%TYPE default null,
        pretty_plural   in rd_columns.pretty_plural%TYPE default null,
        abbrev          in rd_columns.abbrev%TYPE default null,
        description     in rd_columns.description%TYPE default null
    ) return rd_columns.column_id%TYPE;

    procedure edit_column (
        column_id       in rd_columns.column_id%TYPE default null,
        column_name     in rd_columns.column_name%TYPE,
        datatype        in rd_columns.datatype%TYPE,
        default_value   in rd_columns.default_value%TYPE default null,
        attribute_p     in rd_columns.attribute_p%TYPE default 't',
        pretty_name     in rd_columns.pretty_name%TYPE default null,
        pretty_plural   in rd_columns.pretty_plural%TYPE default null,
        abbrev          in rd_columns.abbrev%TYPE default null,
        description     in rd_columns.description%TYPE default null
    );

    procedure drop_column (
        column_id       in rd_columns.column_id%TYPE
    );

    function add_constraint (
        constraint_id   in rd_constraints.constraint_id%TYPE default null,
        column_id       in rd_constraints.column_id%TYPE,
        constraint_type in rd_constraints.constraint_type%TYPE,
        constraint_body in rd_constraints.constraint_body%TYPE default null
    ) return rd_constraints.constraint_id%TYPE;

    procedure drop_constraint (
        constraint_id   in rd_constraints.constraint_id%TYPE
    );

end rapid_development;
/
show errors

create or replace package body rapid_development
is

    function add_table (
        table_id       in rd_tables.table_id%TYPE default null,
        table_name     in rd_tables.table_name%TYPE default null,
        id_column      in rd_tables.id_column%TYPE default null,
        object_type    in rd_tables.object_type%TYPE,
        supertype      in rd_tables.supertype%TYPE default 'acs_object',
        package_name   in rd_tables.package_name%TYPE default null,
        pretty_name    in rd_tables.pretty_name%TYPE,
        pretty_plural  in rd_tables.pretty_plural%TYPE default null,
        abbrev         in rd_tables.abbrev%TYPE default null,
        description    in rd_tables.description%TYPE default null
    ) return rd_tables.table_id%TYPE
    is
        v_table_id      rd_tables.table_id%TYPE;
        v_table_name    rd_tables.table_name%TYPE;
	v_id_column	rd_tables.id_column%TYPE;
        v_package_name  rd_tables.package_name%TYPE;
        v_pretty_plural rd_tables.pretty_plural%TYPE;
    begin
        if table_id is null then
            select rd_table_id_seq.nextval into v_table_id from dual;
        else
            v_table_id := table_id;
        end if;

        if table_name is null then
            v_table_name := object_type || 's';
        else
            v_table_name := table_name;
        end if;

        if id_column is null then
            v_id_column := object_type || '_id';
        else
            v_id_column := id_column;
        end if;

        if package_name is null then
            v_package_name := object_type;
        else
            v_package_name := package_name;
        end if;

        if pretty_plural is null then
            v_pretty_plural := pretty_name || 's';
        else
            v_pretty_plural := pretty_plural;
        end if;

        insert into rd_tables
        (table_id, table_name, id_column, object_type, supertype, package_name,
         pretty_name, pretty_plural, abbrev, description)
        values
        (v_table_id, v_table_name, v_id_column, object_type, supertype, v_package_name,
         pretty_name, v_pretty_plural, abbrev, description);

        return v_table_id;
    end add_table;

    procedure drop_table (
        table_id       in rd_tables.table_id%TYPE
    )
    is
    begin
        delete from rd_tables
        where table_id = drop_table.table_id;
    end drop_table;

    procedure edit_table (
        table_id       in rd_tables.table_id%TYPE,
        table_name     in rd_tables.table_name%TYPE,
        id_column      in rd_tables.id_column%TYPE,
        object_type    in rd_tables.object_type%TYPE,
        supertype      in rd_tables.supertype%TYPE,
        package_name   in rd_tables.package_name%TYPE,
        pretty_name    in rd_tables.pretty_name%TYPE,
        pretty_plural  in rd_tables.pretty_plural%TYPE,
        abbrev         in rd_tables.abbrev%TYPE default null,
        description    in rd_tables.description%TYPE default null
    )
    is
    begin
        update rd_tables
        set    table_name    = edit_table.table_name,
	       id_column     = edit_table.id_column,
               object_type   = edit_table.object_type,
               supertype     = edit_table.supertype,
               package_name  = edit_table.package_name,
               pretty_name   = edit_table.pretty_name,
               pretty_plural = edit_table.pretty_plural,   
               abbrev        = edit_table.abbrev,
               description   = edit_table.description
        where  table_id = edit_table.table_id;
    end edit_table;

    function add_column (
        column_id       in rd_columns.column_id%TYPE default null,
        table_id        in rd_columns.table_id%TYPE,
        column_name     in rd_columns.column_name%TYPE,
        datatype        in rd_columns.datatype%TYPE,
        default_value   in rd_columns.default_value%TYPE default null,
        attribute_p     in rd_columns.attribute_p%TYPE default 't',
        pretty_name     in rd_columns.pretty_name%TYPE default null,
        pretty_plural   in rd_columns.pretty_plural%TYPE default null,
        abbrev          in rd_columns.abbrev%TYPE default null,
        description     in rd_columns.description%TYPE default null
    ) return rd_columns.column_id%TYPE
    is
        v_column_id rd_columns.column_id%TYPE;
    begin
        if column_id is null then
            select rd_column_id_seq.nextval into v_column_id from dual;
        else
            v_column_id := column_id;
        end if;

        insert into rd_columns
        (column_id, table_id, column_name, datatype, default_value,
         attribute_p, pretty_name, pretty_plural, abbrev, description)
        values
        (v_column_id, table_id, column_name, datatype, default_value,
         attribute_p, pretty_name, pretty_plural, abbrev, description);

        return v_column_id;
    end add_column;

    procedure drop_column (
        column_id       in rd_columns.column_id%TYPE
    )
    is
    begin
        delete from rd_columns
        where column_id = drop_column.column_id;
    end drop_column;

    procedure edit_column (
        column_id       in rd_columns.column_id%TYPE default null,
        column_name     in rd_columns.column_name%TYPE,
        datatype        in rd_columns.datatype%TYPE,
        default_value   in rd_columns.default_value%TYPE default null,
        attribute_p     in rd_columns.attribute_p%TYPE default 't',
        pretty_name     in rd_columns.pretty_name%TYPE default null,
        pretty_plural   in rd_columns.pretty_plural%TYPE default null,
        abbrev          in rd_columns.abbrev%TYPE default null,
        description     in rd_columns.description%TYPE default null
    )
    is
    begin
        update rd_columns
        set    column_name   = edit_column.column_name,
               datatype      = edit_column.datatype,
               default_value = edit_column.default_value,
               attribute_p   = edit_column.attribute_p,
               pretty_name   = edit_column.pretty_name,
               pretty_plural = edit_column.pretty_plural, 
               abbrev        = edit_column.abbrev,
               description   = edit_column.description
        where  column_id = edit_column.column_id;
    end edit_column;

    function add_constraint (
        constraint_id   in rd_constraints.constraint_id%TYPE default null,
        column_id       in rd_constraints.column_id%TYPE,
        constraint_type in rd_constraints.constraint_type%TYPE,
        constraint_body in rd_constraints.constraint_body%TYPE default null
    ) return rd_constraints.constraint_id%TYPE
    is
        v_constraint_id rd_constraints.constraint_id%TYPE;
    begin
        if (constraint_body is not null) and (constraint_type in ('primary key', 'unique', 'not null')) then
            raise_application_error(-20000,'Constraint body was not null, but the constraint type ''' || constraint_type || ''' does not accept a body');
        end if;

        if (constraint_body is null) and (constraint_type in ('foreign key', 'check')) then
            raise_application_error(-20000,'The constraint type ''' || constraint_type || ''' requires a body but none was provided');
        end if;

        if constraint_id is null then  
            select rd_constraint_id_seq.nextval into v_constraint_id from dual;
        else
            v_constraint_id := constraint_id;
        end if;

        insert into rd_constraints
        (constraint_id, column_id, constraint_type, constraint_body)
        values
        (v_constraint_id, column_id, constraint_type, constraint_body);

        return v_constraint_id;
    end add_constraint;

    procedure drop_constraint (
        constraint_id   in rd_constraints.constraint_id%TYPE 
    )
    is
    begin
        delete from rd_constraints
        where constraint_id = drop_constraint.constraint_id;
    end drop_constraint;

end rapid_development;
/
show errors
