ad_library {
    Procedures to assist in a rapid prototyping system.

    @author Kevin Scaldeferri
    @creation-date 2000-12-27
    @cvs-id $Id: rapid-development-procs.tcl,v 1.2 2001/06/13 16:32:39 dm Exp $
}

ad_proc -public rd_table_info {
    table_id
} {
    Gets the definition of a table.

    @return an <code>[array get]</code> representation with the following keys:

    <code>table_id, table_name, id_column, object_type, supertype, 
    package_name, pretty_name, pretty_plural, abbrev, description, id_column, 
    columns</code>.

    The value of <code>columns</code> is a TCL list of 
    <code>[array get]</code> representations with these keys:
    <code>column_id, column_name, datatype, default_value, 
    attribute_p, abbrev, description, pretty_name,
    pretty_plural, constraints</code> (<code>id_column</code> 
    is special, and is not included in this list).

    The value of <code>constraints</code> is a TCL list of 
    <code>[array get]</code> representations with these keys:
    <code>constraint_type, constraint_body</code>
} {
    db_1row table {
        select table_id,
               table_name,
	       id_column,
               object_type,
               supertype,
               package_name,
               pretty_name,
               pretty_plural,
               abbrev,
               description
        from   rd_tables
        where  table_id = :table_id
    } -column_array table_def

    set table_def(columns) [list]

    db_foreach columns {
        select column_id,
               column_name,
               datatype,
               default_value,
               attribute_p,
               abbrev,
               description,
               pretty_name,
               pretty_plural
        from   rd_columns
        where  table_id = :table_id
	order by column_name
    } -column_array column_info {
        
        set column_id $column_info(column_id)
        set constraints [list]
        db_foreach constraints {
            select constraint_type,
                   constraint_body
            from   rd_constraints
            where  column_id = :column_id
        } -column_array constraint_info {
            lappend constraints [array get constraint_info]
            if {$constraint_info(constraint_type) == "primary key"} {
                set table_def(id_column) $column_info(column_name)
            }
        }

        set column_info(constraints) $constraints
        lappend table_def(columns) [array get column_info]
    }
               
    return [array get table_def]
}


ad_proc -public rd_generate_create_sql {
    table_def
} {
    Generates an sql file to load the table and associated object_type.
    This procedure is not maximally efficient, as it farms pieces out
    to subprograms which individually loop over the columns, however
    the process is, IMO, much easier to understand this way, and
    this isn't something we expect to do a lot.

} {

    set create_table_sql [rd_generate_create_table_sql $table_def]

    set object_type_plsql [rd_generate_object_type_plsql $table_def]

    set plsql_package [rd_generate_plsql_package $table_def]

    return "
$create_table_sql

$object_type_plsql

$plsql_package

"

}


ad_proc -public rd_generate_drop_sql {
    table_def
} {
    Generates an SQL drop script from the table definition.
} {
    array set table_array $table_def

    set drop_sql "
drop package $table_array(package_name);

begin
    acs_object_type.drop_type('$table_array(object_type)');
end;
/
show errors;

drop table $table_array(table_name);

"

    return $drop_sql
}

ad_proc -public rd_canonical_constraint_name {
    object_type column_name type_abbrev column_abbrev constraint_type
} {
    Constructs a canonical name for a constraint from 
    <code>object_type</code>, <code>column_name</code>, and <code>constraint_type</code>, or, if that is too long,
    <code>type_abbrev</code>, <code>column_abbrev</code>, and <code>constraint_type</code>

    <code>type_abbrev</code> and <code>column_abbrev</code> should be no longer than 13 characters

    <code>constraint_type</code> should be one of: 
    primary key, foreign key, unique, not null, check 
    (or check will be assumed)
} {
    # pick a suffix
    switch $constraint_type {
	{primary key} {
	    set suffix "pk"
	}
	{foreign key} {
	    set suffix "fk"
	}
	{unique} {
	    set suffix "un"
	}
	{not null} {
	    set suffix "nn"
	}
	default {
	    set suffix "ck"
	    # corresponds to check
	}
    }

    set constraint_name "${object_type}_${column_name}_$suffix"
    if {[string length $constraint_name] > 30} {
	# guaranteed to be okay, 13 + 1 + 13 + 1 + 2 = 30
	set constraint_name "${type_abbrev}_${column_abbrev}_$suffix"
    }
    
    return $constraint_name
}

    


ad_proc -public rd_generate_create_table_sql {
    table_def
} {
    Constructs a create table statement from a table definition.
} {
    array set table_array $table_def

    set column_sql_list [list]

    set column_comments ""

    # add special id_column stuff
    
    set column_sql "    $table_array(id_column)"

    # now pad with spaces to 40 columns
    for {set i [string length $column_sql]} {$i < 40} {incr i} {
	append column_sql " "
    }
    
    append column_sql "integer
                                        constraint [rd_canonical_constraint_name $table_array(object_type) $table_array(id_column) $table_array(abbrev) $table_array(id_column) "foreign key"] references acs_objects(object_id) -- fix this!
                                        constraint [rd_canonical_constraint_name $table_array(object_type) $table_array(id_column) $table_array(abbrev) $table_array(id_column) "primary key"] primary key"

    lappend column_sql_list $column_sql
    
    foreach column $table_array(columns) {
        array set column_array $column

        # pad with 4 initial spaces
        set column_sql "    $column_array(column_name)"

        # now pad with spaces to 40 columns
        for {set i [string length $column_sql]} {$i < 40} {incr i} {
            append column_sql " "
        }

        # data type
        if {$column_array(datatype) == "boolean"} {
            append column_sql "char(1)"
        } else {
            append column_sql $column_array(datatype)
        }

        # default value
        if ![empty_string_p $column_array(default_value)] {
            switch -glob $column_array(datatype) {
                "integer" -
                "number*" -
                "date" {
                    append column_sql " default $column_array(default_value)"
                }

                default {
                    #varchar, boolean, clob, blob, other
                    append column_sql " default '[db_quote $column_array(default_value)]'"
                }
            }
        }
        
        # if a boolean, fake it with a check constraint
        if {$column_array(datatype) == "boolean"} {
            set constraint_name "${table_array(table_name)}_${column_array(column_name)}_ck"
            if {[string length $constraint_name] > 30} {
                set constraint_name "${table_array(abbrev)}_${column_array(abbrev)}_ck"
            }
            append column_sql "
                                        constraint $constraint_name
                                        check ($column_array(column_name) in ('t','f'))"
        }

        # add the column_comment
        if ![empty_string_p $column_array(description)] {
            append column_comments "
comment on column $table_array(table_name).$column_array(column_name) is '
$column_array(description)
';
"
        }

        # constraints
        foreach constraint $column_array(constraints) {
            array set constraint_array $constraint

            switch $constraint_array(constraint_type) {
                {primary key} {
                    set constraint_text "primary key"
                }
                {foreign key} {
                    set constraint_text "references $constraint_array(constraint_body)"
                }
                {unique} {
                    set constraint_text "unique"
                }
                {not null} {
                    set constraint_text "not null"
                }
                {check} {
                    set constraint_text "check ($constraint_array(constraint_body))"
                }
            }

	    set constraint_name [rd_canonical_constraint_name $table_array(object_type) $column_array(column_name) $table_array(abbrev) $column_array(abbrev) $constraint_array(constraint_type)]

            # that's 40 spaces before each line.
            append column_sql "
                                        constraint $constraint_name
                                        $constraint_text"
        }

        lappend column_sql_list $column_sql
    }

    # assemble the create table statement

    set create_table_sql "create table $table_array(table_name) (
[join $column_sql_list ",\n"]
);
"

    # tack on the comments

    if ![empty_string_p $table_array(description)] {
        append create_table_sql "
comment on table $table_array(table_name) is '
$table_array(description)
';
"
    }

    append create_table_sql $column_comments

    return $create_table_sql
}

ad_proc -public rd_generate_object_type_plsql {
    table_def
} {
    Constructs the statements to create the object type and attributes
    from a table definition.
} {
    array set table_array $table_def

    set object_type_plsql "
declare
    v_attribute_id acs_attributes.attribute_id%TYPE;
begin
    acs_object_type.create_type (
        object_type => '$table_array(object_type)',
        pretty_name => '[db_quote $table_array(pretty_name)]',
        pretty_plural => '[db_quote $table_array(pretty_plural)]',
        supertype => '$table_array(supertype)',
        table_name => '$table_array(table_name)',
        id_column => '$table_array(id_column)',
        package_name => '$table_array(package_name)'
    );
"

    foreach column $table_array(columns) {
        array set column_array $column

	if {[regexp -nocase {varchar} $column_array(datatype)]} {
	    set datatype "string"	    
        } else {
	    set datatype $column_array(datatype)
	}

        if [string equal $column_array(attribute_p) "t"] {
            append object_type_plsql "
    v_attribute_id := acs_attribute.create_attribute (
        object_type => '$table_array(object_type)',
        attribute_name => '$column_array(column_name)',
        datatype => '$datatype',
        pretty_name => '[db_quote $column_array(pretty_name)]',
        pretty_plural => '[db_quote $column_array(pretty_plural)]',
        default_value => '$column_array(default_value)'
    );
"
#        these aren't necessary, and seem to cause problems
#        table_name => '$table_array(table_name)',
#        column_name => '$column_array(column_name)',

        }

    }

    append object_type_plsql "
end;
/
show errors;"
    return $object_type_plsql
}

ad_proc -public rd_generate_plsql_package {
    table_def
} {
    Constructs the minimal PL/SQL package (new and delete methods) 
    from a table definition.
} {
    array set table_array $table_def

    # build the minimal PL/SQL package

    # start by looping through the columns in the table

    set arg_list [list]
    set insert_columns [list]
    set insert_values  [list]
    set id_column $table_array(id_column)

    # imitating the below, for the id_column. this should be put into a
    # function, or the id column made a real column
    set arg "        $id_column"

    for {set i [string length $arg]} {$i < 40} {incr i} {
	append arg " "
    }

    append arg "in $table_array(table_name).${id_column}%TYPE default null"

    lappend arg_list $arg

    lappend insert_columns $id_column
    lappend insert_values v_$id_column


    foreach column $table_array(columns) {
        array set column_array $column

        set arg "        $column_array(column_name)"

        for {set i [string length $arg]} {$i < 40} {incr i} {
            append arg " "
        }

        append arg "in $table_array(table_name).${column_array(column_name)}%TYPE"
        lappend insert_columns $column_array(column_name)

        if ![empty_string_p $column_array(default_value)] {
            switch -glob $column_array(datatype) {
                "integer" -
                "number*" -
                "date" {
                    append arg " default $column_array(default_value)"
                }

                default {
                    #varchar, boolean, clob, blob, other
                    append arg " default '[db_quote $column_array(default_value)]'"
                }
            }
            lappend insert_values "$column_array(column_name)"
        } else {
            lappend insert_values "$column_array(column_name)"
        }

        lappend arg_list $arg
    }

    # now we have to grab the information about the supertype

    set supertype $table_array(supertype)

    set supertype_id_column [db_string supertype_id_column {
        select id_column
        from   acs_object_types
        where  object_type = :supertype
    } -default ""]
    set supertype_package_name [db_string supertype_package_name {
        select package_name
        from   acs_object_types
        where  object_type = :supertype
    } -default $supertype]

    set supertype_source ""

    db_foreach supertype_source {
        select text
        from   user_source
        where  name = upper(:supertype_package_name)
        and    type = 'PACKAGE'
        order by line
    } {
        append supertype_source $text
    }

    # parse the package definition

    set supertype_arg_list [rd_parse_package_def $supertype_source]

    set supertype_new_args [list]

    foreach arg $supertype_arg_list {
        array set arg_array $arg

        if {$arg_array(argument_name) != $supertype_id_column} {
            set arg_text "        $arg_array(argument_name)"

            for {set i [string length $arg_text]} {$i < 40} {incr i} {
                append arg_text " "
            }

            append arg_text "in $arg_array(argument_type)"
            lappend arg_list $arg_text
            lappend supertype_new_args "            $arg_array(argument_name) => $arg_array(argument_name)"
        } else {
            lappend supertype_new_args "            $arg_array(argument_name) => $id_column"
        }
    }

    set plsql_package "
create or replace package $table_array(package_name)
is

    function new (
[join $arg_list ",\n"]
    ) return $table_array(table_name).${table_array(id_column)}%TYPE;

    procedure delete (
        $table_array(id_column) in $table_array(table_name).${table_array(id_column)}%TYPE
    );
end $table_array(package_name);
/
show errors;

create or replace package body $table_array(package_name)
is

    function new (
[join $arg_list ",\n"]
    ) return $table_array(table_name).${table_array(id_column)}%TYPE
    is
        v_$table_array(id_column) $table_array(table_name).${table_array(id_column)}%TYPE;
    begin
        v_$table_array(id_column) := ${supertype_package_name}.new (
[join $supertype_new_args ",\n"]
        );

        insert into $table_array(table_name)
        ([join $insert_columns ", "])
        values
        ([join $insert_values ", "]);

        return v_$table_array(id_column);
    end new;

    procedure delete (
        $table_array(id_column) in $table_array(table_name).${table_array(id_column)}%TYPE
    )
    is
    begin
        delete from $table_array(table_name)
        where  $table_array(id_column) = ${table_array(package_name)}.delete.$table_array(id_column);

        ${supertype_package_name}.delete($table_array(id_column));
    end delete;
end $table_array(package_name);
/
show errors;"

    return $plsql_package
}


ad_proc -private rd_parse_package_def {
    source
} {
    Parses a PL/SQL package definition for the signature of the
    new function.  Returns a list of <code>array get</code> 
    representations with the keys <code>argument_name</code> and 
    <code>argument_type</code>
} {

    # get the new function
    regexp {function +new[^;]*;} $source new_source

    # just grab the args
    regexp {\((.*)\)} $new_source match new_args

    # strip any comments (note that we'd better be using a TCL
    # with non-greedy matching)
    regsub -all {/\*.*?\*/} $new_args {} new_args
    regsub -all -- {--[^\n]*}  $new_args {} new_args

    # split on commas
    set new_args_list [split $new_args ","]

    # now parse out the argument name and type
    # Note that we are assuming all new functions use only in variables

    set args [list]

    foreach arg $new_args_list {
        set arg [string trim $arg]
        regexp {^(\w+)\s+in\s+(.+)$} $arg match name type

        set argument(argument_name) $name
        regsub {\s+} $type { } argument(argument_type)

        lappend args [array get argument]
    }

    return $args
}